/*
   This program tests the communications channels in a parallel computer
   to see if they have similar speeds.

   You'd think that vendors would have such a diagnostic; you'd be wrong.

   The method is to pass a token between pairs of processors, for all 
   immediage neighbors.  This tests ONLY neighbor links; it does not
   test pass-through effects (so, for a circuit-switched or packet-switched
   system, not all routes are tested).  The times are compared; routes
   whose times very greatly from the average are flagged.

   Recently changed to use a device-independent graphics display
 */

#include "tools.h"
#include "comm/comm.h"
#include "system/time/usec.h"
#include <math.h>
#include <stdio.h>
#include "dgraph/dgraph.h"
#if defined(intelnx)
#include <sys/types.h>
#if !defined(intelparagon)
#include <nxinfo.h>
#define PHYNODE() nxinfo.physnode
#else
#define PHYNODE() PImytid
#endif
#else /* Not intel */
#define PHYNODE() PImytid
#endif

/* Forward references */
void OrderNbrs(), TokenTestSync(), TokenTestASync(), TokenTestForce(),
     GenerateReport(), DrawHistogram(), DrawDanceHall(), DrawMesh(), 
     DrawGRHistogram();
double RemoveOutliers();
int *CollectData();

main(argc,argv)
int  argc;
char **argv;
{
int worker();
PICall( worker, argc, argv );
}

#define MAX_NBRS 1024
static GRctx        *gctx = 0;

int worker( argc, argv )
int  argc;
char **argv;
{
double       *times;
int          nnbrs, nbrs[MAX_NBRS], mtype, badnbrs[MAX_NBRS];
int          len, reps, nbr, sval[3], ctype, myid;
double       rtol;
int          k, do_graph = 0, nx = 0, ndim[2];

/* Get test parameters */
if (SYArgHasName( &argc, argv, 1, "-help" )) {
    if (PImytid != 0) return;
    fprintf( stderr,
    "%s -size len max incr", argv[0] );
    fprintf( stderr, "\
-rtol <tolerance> -reps <repititions> -async -force\n\
-graph      Draw graph of communication test\n\
-nx n\n\
-mesh nx ny Use 2d mesh topology\n\
-all        Use complete connection topology\n" );
    return;
    }
    
myid    = PImytid;
rtol    = 0.05;
reps    = 100;
ctype   = 0;
sval[0] = 64;
sval[1] = 64;
sval[2] = 64;
SYArgGetDouble( &argc, argv, 1, "-rtol", &rtol );
SYArgGetInt(    &argc, argv, 1, "-reps", &reps );
SYArgGetIntVec( &argc, argv, 1, "-size", 3, sval );
if (SYArgHasName( &argc, argv, 1, "-async" )) ctype = 1;
if (SYArgHasName( &argc, argv, 1, "-force" )) ctype = 2;
SYArgGetInt( &argc, argv, 1, "-nx", &nx );
do_graph = SYArgHasName( &argc, argv, 1, "-graph" );
if (do_graph && PImytid == 0) {
    /* Use "-grfname name" for output graphics file */
    gctx = GRCreateFromArgsNoX( &argc, argv );
    }

/* Get the neighbors */
if (SYArgHasName( &argc, argv, 1, "-all" )) {
    nnbrs = 0;
    for (k=0; k<PInumtids; k++) {
	if (k != PImytid)
	    nbrs[nnbrs++] = k;
        }
    }
else if (SYArgGetIntVec( &argc, argv, 1, "-mesh", 2, ndim )) {
    nx = ndim[0];
    }
else
    nnbrs = PIGetNbrs( PImytid, nbrs );

/* need to compute the message type to use and order so that nodes don't
   block.  Note that everyone already knows all the processors that they
   will be communicating with. */
OrderNbrs( nnbrs, nbrs );

times = (double *)MALLOC( nnbrs * sizeof(double) );   CHKPTRV(times,1);

for (len=sval[0]; len<=sval[1]; len+=sval[2]) {
    /* For each neighbor, start the test.  
     */
    for (nbr = 0; nbr < nnbrs; nbr++) {
        mtype = (nbrs[nbr] > myid) ? myid : nbrs[nbr];
        switch (ctype) {
        case 0:
            TokenTestSync( nbrs[nbr], times + nbr, mtype, len, reps );  break;
        case 1:
            TokenTestASync( nbrs[nbr], times + nbr, mtype, len, reps ); break;
        case 2:
            TokenTestForce( nbrs[nbr], times + nbr, mtype, len, reps ); break;
            }
	}
    /* Generate report */
    GenerateReport( nbrs, nnbrs, times, rtol, len, reps, badnbrs, do_graph );
    if (do_graph) {
	if (nx == 0) 
	    DrawDanceHall( gctx, nbrs, nnbrs, times, rtol, badnbrs );
	else 
	    DrawMesh( gctx, nbrs, nnbrs, times, rtol, badnbrs, nx );
	}
    }
    
FREE( times );
/* This won't work for batch jobs (like EUI-H...) */
if (do_graph && PImytid == 0) {
    int c;
    if (GRIsInteractive(gctx)) {
#ifdef euih
	sleep( 20 );
#else
	printf( "Press return to end program..." );
	c = getc( stdin );
#endif
	}
    GRClose( gctx );
    }
}

/* To get the timing right, the low processor sends to the high processor
   to start the test, then the timing starts */
void TokenTestSync( nbr, time, phase, len, reps )
int    nbr, phase, len, reps;
double *time;
{
int          myid, i;
SYusc_time_t t1, t2;
char         *sbuf, *rbuf;

myid = PImytid;
MSGALLOCRECV(rbuf,len,char);
MSGALLOCSEND(sbuf,len,char);

if (myid < nbr) {
    SENDSYNC( phase, sbuf, 0, nbr, MSG_OTHER );
    RECVSYNC( phase, rbuf, len, MSG_OTHER );
    SYusc_clock( &t1 );
    for (i=0; i<reps; i++) {
	SENDSYNC( phase, sbuf, len, nbr, MSG_OTHER );
	RECVSYNC( phase, rbuf, len, MSG_OTHER );
	}
    SYusc_clock( &t2 );
    *time = SYuscDiff( &t1, &t2 );
    }
else {
    RECVSYNC( phase, rbuf, len, MSG_OTHER );
    SENDSYNC( phase, sbuf, len, nbr, MSG_OTHER );
    SYusc_clock( &t1 );
    for (i=0; i<reps; i++) {
	RECVSYNC( phase, rbuf, len, MSG_OTHER );
	SENDSYNC( phase, sbuf, len, nbr, MSG_OTHER );
	}
    SYusc_clock( &t2 );
    *time = SYuscDiff( &t1, &t2 );
    }
MSGFREERECV(rbuf);
MSGFREESEND(sbuf);
}

void TokenTestASync( nbr, time, phase, len, reps )
int    nbr, phase, len, reps;
double *time;
{
int            myid, i;
SYusc_time_t   t1, t2;
char           *sbuf, *rbuf;
ASYNCRecvId_t  rid;

myid = PImytid;
MSGALLOCRECV(rbuf,len,char);
MSGALLOCSEND(sbuf,len,char);

if (myid < nbr) {
    SENDSYNC( phase, sbuf, 0, nbr, MSG_OTHER );
    RECVSYNC( phase, rbuf, len, MSG_OTHER );
    SYusc_clock( &t1 );
    for (i=0; i<reps; i++) {
    	RECVASYNC( phase, rbuf, len, MSG_OTHER, rid );
	SENDSYNC( phase, sbuf, len, nbr, MSG_OTHER );
	RECVWAIT( phase, rbuf, len, MSG_OTHER, rid );
	}
    SYusc_clock( &t2 );
    *time = SYuscDiff( &t1, &t2 );
    }
else {
    RECVSYNC( phase, rbuf, len, MSG_OTHER );
    RECVASYNC( phase, rbuf, len, MSG_OTHER, rid );
    SENDSYNC( phase, sbuf, len, nbr, MSG_OTHER );
    SYusc_clock( &t1 );
    for (i=0; i<reps-1; i++) {
	RECVWAIT( phase, rbuf, len, MSG_OTHER, rid );
	SENDSYNC( phase, sbuf, len, nbr, MSG_OTHER );
	RECVASYNC( phase, rbuf, len, MSG_OTHER, rid );
	}
    RECVWAIT( phase, rbuf, len, MSG_OTHER, rid );
    SENDSYNC( phase, sbuf, len, nbr, MSG_OTHER );
    SYusc_clock( &t2 );
    *time = SYuscDiff( &t1, &t2 );
    }
MSGFREERECV(rbuf);
MSGFREESEND(sbuf);
}

/* Use async code once it works, but with force modifier */
void TokenTestForce( nbr, time, phase, len, reps )
int    nbr, phase, len, reps;
double *time;
{
}

#define MAXPHYLEN 40
typedef struct {
    int id, partner;
    double time;
    } BadList;

/*
   This routine analyzes the results, looking for unusually fast or
   slow nodes.  A histogram of the times is produced as well.

   badnbrs[i] = +/- 1 if nbrs[i] is out-of-range, 0 otherwise.
                (+ if time > average, - if < average)
 */
void GenerateReport( nbrs, nnbrs, times, rtol, len, reps, badnbrs, do_graph )
int    *nbrs, nnbrs, len, reps, *badnbrs, do_graph;
double *times, rtol;
{
int     i, nlinks, j;
int     cnt = 0;
double  rlow, rhigh;
double  mintime, maxtime, avetime, wtime;
char    phy_name[MAXPHYLEN];
int     nslow = 0;        /* This is a simple list of the 5 slowest links */
BadList slowest[5];
int     kk, jj;

/* Get some information on the global times.  We are assuming that
   all neighbors should have the same speed of links */
maxtime = times[0];
mintime = times[0];
avetime = times[0];
for (i=1; i<nnbrs; i++) {
    avetime += times[i];
    if (times[i] > maxtime)      maxtime = times[i];
    else if (times[i] < mintime) mintime = times[i];
    }

PIgdmax( &maxtime, 1, &wtime, ALLPROCS );
PIgdmin( &mintime, 1, &wtime, ALLPROCS );
PIgdsum( &avetime, 1, &wtime, ALLPROCS );
nlinks  = nnbrs;
PIgisum( &nlinks, 1, &i, ALLPROCS );
avetime /= nlinks;

for (i=0; i<nnbrs; i++) 
    badnbrs[i] = 0;
/*
   Here we could try to remove outliers and recompute the average
   (that is, discard any local values that are well out-of-range, then
   recompute the average time on the remaining links.  Do this until
   no further values are discarded.
*/

/* Save the name of THIS processor */
PINodeName( phy_name, MAXPHYLEN );

/* Look for nodes that are away from the mean */
if (maxtime - mintime >= rtol * avetime) {
    /* Compute a new average */
    avetime = RemoveOutliers( times, nnbrs, 
			      mintime, maxtime, 2*rtol, avetime );
    
    /* Somebody is bad */
    rlow  = avetime * (1.0 - rtol);
    rhigh = avetime * (1.0 + rtol);
    if (PImytid == 0) 
      printf( "%cNode[PhysNode] Nbr         Time  AverageTime        %%Diff\n",
	      do_graph ? '#' : ' ' );

    for (j=0; j<=PInumtids; j++) {
    	/* If we could pass a data-value with the token, we could
    	   pass the error count along.  If we were the first one,
    	   we could issue an error message. */
    	if (PIgtoken(PSAllProcs,j)) {
            for (i = 0; i < nnbrs; i++) {
                if (times[i] < rlow || times[i] > rhigh) {
                    cnt ++;
             	    printf( "%c%4d[%12s] %3d %12.2e %12.2e %12.2e\n", 
			    do_graph ? '#' : ' ', 
			    j, phy_name, nbrs[i], times[i], avetime, 
	                    100.0*(times[i] - avetime)/avetime );
		    fflush( stdout );
		    badnbrs[i] = (times[i] < rlow) ? -1 : 1;
	            }
		/* Update the list of slowest links */
		if (nslow < 5) {
		    /* See if we should insert it into the middle ... */
		    for (kk=0; kk<nslow; kk++) {
			if (slowest[kk].time < times[i]) {
			    /* Move the values down */
			    for (jj=nslow; jj>kk; jj--) 
				slowest[jj] = slowest[jj-1];
			    slowest[kk].id      = j;
			    slowest[kk].partner = nbrs[i];
			    slowest[kk].time    = times[i];
			    break;
			    }
			}
		    if (kk == nslow) {
			slowest[kk].id = j;
			slowest[kk].partner = nbrs[i];
			slowest[kk].time = times[i];
			}
		    nslow++;
		    }
		else if (times[i] > slowest[nslow-1].time) {
		    for (kk=0; kk<nslow; kk++) {
			if (slowest[kk].time < times[i]) {
			    /* Move the values down */
			    for (jj=4; jj>kk; jj--) 
				slowest[jj] = slowest[jj-1];
			    slowest[kk].id      = j;
			    slowest[kk].partner = nbrs[i];
			    slowest[kk].time    = times[i];
			    break;
			    }
			}
		    }
                }
            }
        }
    }
    
PIgisum(&cnt,1,&i,ALLPROCS);
if (PImytid == 0) {
    if (do_graph) printf( "#" );
    if (cnt == 0) printf( "All links within range\n" );
    else          printf( "%d links are out-of-range\n", cnt );
    if (do_graph) printf( "#" );
    printf( "For message size = %d, Average rate = %.0f bytes/sec\n",
	    len, 2.0 * (double)(len * reps) / avetime );
    }
if (cnt) {
    int nbin = 40;
    while (nbin > nlinks * 2) nbin /= 2;
    if (PImytid == 0) 
	printf( "\n%cHistogram by time on each link\n", do_graph ? '#' : ' ' );
    DrawHistogram( times, nnbrs, nbin, stdout, ALLPROCS, mintime, maxtime, 
		   do_graph );
    if (do_graph)
	DrawGRHistogram( gctx, times, nnbrs, PIAllProcs, mintime, maxtime );
    }

/* Data for a plot of processors versus time would also be interesting,
   perhaps to an auxillery file */

/* Also useful to do a short list of the physical nodes involved in the 
   SLOWEST links, the help assimilate the information */
#ifdef FOO
still need to get the top 5 over the machine... 
for (j=0; j<=PInumtids; j++) {
    if (PIgtoken(PSAllProcs,j)) {
	for (kk=0; kk<nslow; kk) {
	    printf( "[%d] 
	    }
	}
    }
#endif
}


/* 
   Algorithm for ordering the partners.
   This is an ordering that can be computed entirely locally for
   each processor.  A similar ordering is used in BlockComm for
   synchronous communication.

   The idea is to break the communication up into phases between processors
   that differ in a current bit position (mask).  Only those processors
   may communicate in that phase.

   This algorithm is NOT guarenteed to generate an optimial ordering.
   It does do so for hypercubes and for meshes with even dimension.
   Meshes with odd dimension will have a poor schedule (roughly
   proportional to the diameter of the mesh).
 */

void OrderNbrs( nnbrs, nbrs )
int  nnbrs, *nbrs;
{
int mask = 0x1, i;
int *newnbrs, cnt, myid;

myid = PImytid;
newnbrs = (int *)MALLOC( nnbrs * sizeof(int) );    CHKPTR(newnbrs);

/* Sort by increasing node number */
SYIsort( nnbrs, nbrs );

cnt     = 0;
while (cnt < nnbrs) {
    for (i=0; i<nnbrs; i++) {
	if (nbrs[i] >= 0 && ((myid & mask) ^ (nbrs[i] & mask)) &&
	    myid > nbrs[i]) {
	    /* Masters */
	    newnbrs[cnt++] = nbrs[i];
	    nbrs[i] = -1;
	    }
	}
    for (i=0; i<nnbrs; i++) {
	if (nbrs[i] >= 0 && ((myid & mask) ^ (nbrs[i] & mask))) {
	    /* Slaves */
	    newnbrs[cnt++] = nbrs[i];
	    nbrs[i] = -1;
	    }
	}
    mask <<= 1;
    }

MEMCPY(nbrs,newnbrs,nnbrs*sizeof(int));
FREE(newnbrs);
}

/*
    Draw a histogram on FILE fp using (double) data and nbin bins
 */     
void DrawHistogram( data, n, nbin, fp, procset, dmin, dmax, do_graph )
double  *data, dmin, dmax;
int     n, nbin, do_graph;
FILE    *fp;
ProcSet *procset;
{
char *line;
int  *bins, *work;
int  i, j, ib;
int  maxcnt;

bins = (int *)MALLOC( nbin * 2 * sizeof(int) );   CHKPTR(bins);
work = bins + nbin;
line = (char *)MALLOC( nbin + 1 );                CHKPTR(line);

for (i=0; i<nbin; i++)
    bins[i] = 0;

for (i=0; i<n; i++) {
    ib  = (nbin - 1) * (data[i] - dmin) / (dmax - dmin);
    bins[ib]++;
    }
PIgisum( bins, nbin, work, procset );

if (PSISROOT(procset)) {
    maxcnt = 0;
    for (i=0; i<nbin; i++) {
        line[i] = ' ';
        if (bins[i] > maxcnt) maxcnt = bins[i];
        }
    line[nbin] = 0;
    for (j=maxcnt; j>0; j--) {
    	for (i=0; i<nbin; i++)
    	    if (bins[i] == j) line[i] = '*';
    	fprintf( fp, "%c%s\n", do_graph ? '#' : ' ', line );
        }
    fprintf( fp, "%cmin = %12.2e max = %12.2e\n", 
	     do_graph ? '#' : ' ', dmin, dmax );
    }
FREE( bins );
FREE( line );
}

/* Compute a new average time after discarding times that are away
   from the average.

   This should really be enhanced to remove any far-out clumps.  For example,
   if the histogram is

   0 10 9 0 0 0 0 0 0 0 0 2 0 0 2

   The 4 on the end should be considered outliers.  One possibility
   is to accept any negative number bu reject large positive differences;
   this would eliminate these.

   Another possibility is to successively remove the fartherest from the 
   average until the remainer fits within the original limits.  We could
   even do a fit to the data, and remove the outside ones until the remainder
   had a good fit.
   
*/
double RemoveOutliers( times, n, mintime, maxtime, rtol, avetime )
double *times, mintime, maxtime, rtol;
int    n;
{
int    cnt, i;
double sum, work;
double newave;
int    *discard;   /* indicates if a value has been discarded */
int    maxdev;     /* index of maximum deviation */
double curdev;     /* size of current deviation */
double glodev;     /* global deviation */
double sum2;       /* used to compute new average */
int    cnt2; 

if (PImytid == 0) {
    printf("About to try and remove outliers.  This may take some time...\n" );
    fflush( stdout );
    }
discard = (int *)MALLOC( n * sizeof(int) );
for (i=0; i<n; i++) discard[i] = 0;
cnt    = 0;
newave = avetime;
while (cnt == 0) {
    cnt    = 0;
    sum    = 0.0;
    maxdev = -1;
    curdev = 0.0;
    for (i=0; i<n; i++) {
	if (!discard[i] && fabs( times[i] - newave ) > rtol * newave) {
	    /* Keep track of the largest deviation */
	    if (maxdev == -1 || fabs( times[i] - newave ) > curdev) {
		maxdev = i;
		curdev = fabs( times[i] - newave );
		}
	    }
	if (fabs( times[i] - newave ) < rtol * newave ) {
	    cnt++;
	    sum += times[i];
	    }
	}

    glodev = curdev;
    PIgdmax( &glodev, 1, &work, PIAllProcs );
    if (maxdev >= 0 && curdev >= 0.95 * glodev)  {
	/* printf( "Discarding outlier (%d)\n", PImytid );
	fflush( stdout ); */
	discard[maxdev] = 1;
	}
    /* compute a new average */
    sum2 = 0.0;
    cnt2 = 0;
    for (i=0; i<n; i++) {
	if (!discard[i]) {
	    sum2 += times[i];
	    cnt2++;
	    }
	}
    PIgdsum( &sum2, 1, &work, PIAllProcs );
    PIgisum( &cnt2, 1, &i,    PIAllProcs );
    if (cnt2 > 0) newave = sum2 / cnt2;
    /* printf( "New average is %f\n", newave ); */
	    
    PIgdsum( &sum, 1, &work, ALLPROCS );
    PIgisum( &cnt, 1, &i,    ALLPROCS );
    if (glodev == 0.0) {
	/* We didn't find anyone to remove, so we'll expand the limits */
	/* If cnt is STILL 0, EVERYTHING was an outlier, so we prepare to 
	   try again with twice the tolerance */
	rtol *= 2.0;
	}
    }
FREE( discard );
return sum / cnt;
}

/*
   The following code (not yet written) 
   is a first attempt to help generate graphical 
   output for the output from this routine.  The idea is

   If the topology is known and has limited connectivity (i.e., mesh), 
   then draw it.  For this, we need to know our location in the mesh.

   If the topology is not known or is highly connected, generate one of 
   two plots:
       for each node, a separate plot of all connections
       a "dance hall" diagram.

   The graphs should draw in-range lines as dotted and out-of-range lines
   as solids

   Note that since the data is distributed, these are parallel routines.
 */

/* type == 0 for in range, 1 otherwise */
void DrawConnection( gctx, fromx, fromy, tox, toy, type )
GRctx *gctx;
int  fromx, fromy, tox, toy, type;
{
GRStyle style;

if (type == 0) {
    style.line_style = GRSOLID;
    style.color      = GRBLACK;
    }
else if (type < 0) {
    style.line_style = GRDOTS;
    style.color      = GRGREEN;
    }
else {
    style.line_style = GRDASH;
    style.color      = GRRED;
    }
GRDrawLine( gctx, (double)fromx, (double)fromy, (double)tox, (double)toy,
	    &style );
}

void DrawNode( gctx, xc, yc, num )
GRctx *gctx;
int  xc, yc, num;
{
char snum[5];	
GRStyle style;
style.line_style = GRSOLID;
style.color      = GRBLACK;
GRDrawRectangle( gctx, (double) (xc - 2), (double) (xc + 2),
		       (double) (yc - 2), (double) (yc + 2), &style );
sprintf( snum, "%d", num );
GRDrawText( gctx, (double)xc, (double)yc, snum, 10, GRCENTER ); 
}

/* 
   Draw a dance-hall diagram (all processors on bottom and top)

   This uses a single program (the root) to draw, avoiding 
   problems with multiple writers and connections
 */
void DrawDanceHall( gctx, nbrs, nnbrs, times, rtol, badnbrs )
GRctx  *gctx;
int    *nbrs, nnbrs, *badnbrs;
double *times, rtol;
{
int i, j, k, *icol;

if (PImytid == 0) {
    GRStartPage( gctx );	
    GRSetLimits( gctx, -3.0, (double)(PInumtids * 5), -1.0, 15.0 );
    GRDrawAxis( gctx, 0 );
    }
/* The approach is for each node to wait until node 0 tells them to go ahead,
   then they send a single record of the form nnbrs(to,link)*; 
   To speed the collection, PIgcol is used */

/* Draw the nodes */
if (PImytid == 0) {
    for (i=0; i<PInumtids; i++) {
        DrawNode( gctx, i*5, 2,  i );
        DrawNode( gctx, i*5, 12, i );
        }
    }
icol = CollectData( nnbrs, nbrs, badnbrs );
if (PImytid == 0) {
    for (i=0; i<PInumtids; i++) {
        nnbrs = icol[0];
	j     = icol[1];   /* Source */
	for (k=0; k<nnbrs; k++) {
	    DrawConnection( gctx, j*5, 4, icol[2*k+2]*5, 10, icol[2*k+3] );
	    }
	icol += icol[0] * 2 + 2;
	}
    GREndPage( gctx );    
    }
}

void DrawMesh( gctx, nbrs, nnbrs, times, rtol, badnbrs, nx )
GRctx  *gctx;
int    *nbrs, nnbrs, *badnbrs, nx;
double *times, rtol;
{
int i, k, ifrom, jfrom, ito, jto, *icol;

if (PImytid == 0) {
    GRSetLimits( gctx, -3.0, (double)(nx * 5), 
		       -3.0, (double)((PInumtids / nx) * 5 - 2) );
    GRDrawAxis( gctx, 0 );
    }
/* Draw the nodes in a mesh framework */
if (PImytid == 0) {
    for (i=0; i<=PInumtids; i++) {
	ifrom = i % nx;
	jfrom = i / nx;
	DrawNode( gctx, ifrom*5, jfrom*5,  i );
        }
    }
icol = CollectData( nnbrs, nbrs, badnbrs );

if (PImytid == 0) {
    for (i=0; i<PInumtids; i++) {
    	ifrom = icol[1] % nx;
	jfrom = icol[1] / nx;
	nnbrs = icol[0];
	for (k=0; k<nnbrs; k++) {
	    ito = icol[2*k+2] % nx;
	    jto = icol[2*k+2] / nx;
	    if (jfrom == jto) {
		if (ifrom > ito) 
		    DrawConnection( gctx, ito*5+2, jto*5 - 1, 
				    ifrom*5-2, jfrom*5 - 1, icol[2*k+3] );
		else
		    DrawConnection( gctx, ifrom*5+2, jfrom*5 + 1, ito*5-2, 
				    jto*5 + 1,  icol[2*k+3] );
		}
	    else if (jfrom < jto) 
		DrawConnection( gctx, ifrom*5+1, jfrom*5 + 2, 
			        ito*5+1, jto*5 - 2, icol[2*k+3] );
	    else 
		DrawConnection( gctx, ito*5 - 1, jto*5 + 2, 
			        ifrom*5 - 1, jfrom*5 - 2, icol[2*k+3] );
	    }
	icol += icol[0] * 2 + 2;
	}
    }
}

void DrawOneToAll( )
{
}

int *CollectData( nnbrs, nbrs, badnbrs )
int nnbrs, *nbrs, *badnbrs;
{
int *icol, *lcol, k, nlinks, iwork, glen;

/* Get the number of links */
nlinks = nnbrs;
PIgisum( &nlinks, 1, &iwork, PSAllProcs );

/* Allocate storage */
icol = (int *)MALLOC( (nlinks * 2 + 2 * PInumtids) * sizeof(int) );  
lcol = (int *)MALLOC( (nlinks * 2 + 2 * PInumtids) * sizeof(int) );  
CHKPTRN(icol);
CHKPTRN(lcol);
lcol[0] = nnbrs;
lcol[1] = PImytid;
for (k=0; k<nnbrs; k++) {
    lcol[2*k+2] = nbrs[k];
    lcol[2*k+3] = badnbrs[k];
    }
PIgcol( lcol, (2*nnbrs+2)*sizeof(int), icol, 
        (nlinks * 2 + 2 * PInumtids)*sizeof(int), &glen, PSAllProcs, MSG_INT );
FREE(lcol);
return icol;
}

/*
  Draw a histogram using GR code.  Only ONE GR device is open, so 
  data must be sent to the master node
 */     
void DrawGRHistogram( gctx, data, n, procset, dmin, dmax )
GRctx   *gctx;
double  *data, dmin, dmax;
int     n;
ProcSet *procset;
{
double *gdata;
int    nnbrs, i, j, off;

nnbrs = n;
PIgisum( &nnbrs, 1, &i, procset );
if (PImytid > 0) {
    for (i=0; i<=PInumtids; i++) {
	if (PIgtoken( procset, i )) {
	    PIbsend( 11, data, n * sizeof(double), 0, MSG_DBL );
	    }
	}
    }
else {
    gdata = (double *)MALLOC( nnbrs * sizeof(double) );
    off   = 0;
    for (i=0; i<=PInumtids; i++) {
	PIgtoken( procset, i );
	if (i > 0 && i < PInumtids) {
	    PIbrecv( 11, gdata+off, (nnbrs-off) * sizeof(double), MSG_DBL );
	    off += PIsize() / sizeof(double);
	    }
	else if (i == 0) {
	    for (j=0; j<n; j++)
		gdata[j] = data[j];
	    off += n;
	    }
	}
    GRStartPage( gctx );	
    GRDrawHistogram( gctx, gdata, nnbrs, (dmax - dmin) / 40 );
    GREndPage( gctx );    
    FREE( gdata );
    }
}
